import pytest
from tests.models_to_test import ModelsToTest
import asyncio
from sl_ai_models.basic_model_interfaces.priced_per_request import PricedPerRequest
from sl_ai_models.basic_model_interfaces.tokens_incur_cost import TokensIncurCost
from sl_ai_models.utils.response_types import TextTokenCostResponse
from sl_ai_models.basic_model_interfaces.ai_model import AiModel
from sl_ai_models.model_archetypes.traditional_online_llm import TraditionalOnlineLlm

@pytest.mark.parametrize(
    "subclass",
    ModelsToTest.TOKENS_INCUR_COST_LIST
)
def test_predicted_cost_and_tokens_correct_for_cheap_input(subclass: type[TokensIncurCost]) -> None:
    model = subclass()
    assert_cost_and_tokens_correct_for_cheap_input(model)


@pytest.mark.parametrize(
    "subclass",
    ModelsToTest.TOKENS_INCUR_COST_LIST
)
def test_system_prompt_cost_and_tokens_correct_for_cheap_input(subclass: type[TokensIncurCost]) -> None:
    if not issubclass(subclass, TraditionalOnlineLlm):
        pytest.skip("Model doesn't have a system prompt")

    model_without_system_prompt = subclass()
    model_with_system_prompt = subclass(system_prompt="This is a system prompt")
    cheap_input = model_without_system_prompt._get_cheap_input_for_invoke()

    with_system_prompt_tokens = model_with_system_prompt.input_to_tokens(cheap_input)
    model_without_system_prompt_tokens = model_without_system_prompt.input_to_tokens(cheap_input)
    assert with_system_prompt_tokens != model_without_system_prompt_tokens, "Token cost should differ if system prompt is set"

    assert_cost_and_tokens_correct_for_cheap_input(model_with_system_prompt)



def assert_cost_and_tokens_correct_for_cheap_input(model: TokensIncurCost) -> None:
    assert isinstance(model, AiModel), "model must be an instance of AiModel"

    cheap_input = model._get_cheap_input_for_invoke()

    predicted_prompt_tokens = model.input_to_tokens(cheap_input)
    predicted_prompt_cost_v1 = model.caculate_cost_from_tokens(predicted_prompt_tokens,0)
    predicted_prompt_cost_v2 = model.cost_per_token_prompt * predicted_prompt_tokens

    response = asyncio.run(model._mockable_direct_call_to_model(cheap_input))

    if not isinstance(response, TextTokenCostResponse):
        raise ValueError("response must be a TextTokenCostResponse or the test will fail")

    actual_prompt_tokens = response.prompt_tokens_used
    actual_completion_tokens = response.completion_tokens_used
    actual_cost = response.cost

    predicted_completion_cost_v1 = model.caculate_cost_from_tokens(0,actual_completion_tokens)
    if isinstance(model, PricedPerRequest):
        predicted_completion_cost_v1 -= model.PRICE_PER_REQUEST
    predicted_completion_cost_v2 = model.cost_per_token_completion * actual_completion_tokens

    predicted_total_cost_v1 = predicted_prompt_cost_v1 + predicted_completion_cost_v1
    predicted_total_cost_v2 = predicted_prompt_cost_v2 + predicted_completion_cost_v2

    allowed_error = 0.0001
    assert predicted_prompt_tokens == actual_prompt_tokens, f"predicted tokens different than actual tokens. Predicted: {predicted_prompt_tokens}, Actual: {actual_prompt_tokens}"
    assert abs(predicted_total_cost_v1 - actual_cost) < allowed_error, f"predicted cost different than actual cost using method 1. Predicted: {predicted_prompt_cost_v1}, Actual: {actual_cost}"

    predicted_cost_v2_within_allowed_error = abs(predicted_total_cost_v2 - actual_cost) < allowed_error
    if not isinstance(model, PricedPerRequest):
        assert predicted_cost_v2_within_allowed_error, f"predicted cost different than actual cost using method 2. Predicted: {predicted_prompt_cost_v2}, Actual: {actual_cost}"
    else:
        assert not predicted_cost_v2_within_allowed_error, "Price per request should have broken this"


from __future__ import annotations
from abc import ABC


class PricedPerRequest(ABC):
    PRICE_PER_REQUEST: float = NotImplemented

    def __init_subclass__(cls: type[PricedPerRequest], **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        if ABC not in cls.__bases__ and cls.PRICE_PER_REQUEST is NotImplemented:
            raise NotImplementedError('You forgot to define PRICE_PER_REQUEST')



from sl_ai_models.basic_model_interfaces.incurs_cost import IncursCost
from abc import ABC, abstractmethod
from sl_ai_models.basic_model_interfaces.tokens_are_calculatable import TokensAreCalculatable
from sl_ai_models.resource_managers.monetary_cost_manager import MonetaryCostManager, UsagePrediction
from sl_ai_models.utils.response_types import TextTokenCostResponse
from typing import Any


class TokensIncurCost(TokensAreCalculatable, IncursCost, ABC):

    @abstractmethod
    def caculate_cost_from_tokens(self, prompt_tkns: int, completion_tkns: int) -> float:
        pass


    @abstractmethod
    def input_to_tokens(self, *args, **kwargs) -> int:
        pass


    def _estimate_before_plus_after_cost_from_model_input(self, *args, **kwargs) -> float:
        prompt_tokens = self.input_to_tokens(*args, **kwargs)
        heuristic_multiplier_for_completion_tokens = 2.5
        estimated_completion_tokens = int(prompt_tokens * heuristic_multiplier_for_completion_tokens)
        estimated_cost = self.caculate_cost_from_tokens(prompt_tkns=prompt_tokens, completion_tkns=estimated_completion_tokens)
        return estimated_cost


    async def _track_cost_in_manager_using_model_response(self, response_from_direct_call: Any, original_prediction: UsagePrediction) -> None:
        if isinstance(response_from_direct_call, TextTokenCostResponse):
            cost = response_from_direct_call.cost
        else:
            raise NotImplementedError(f"This method has not been implemented for response type {type(response_from_direct_call)}")
        MonetaryCostManager.replace_predicted_usage_with_actual_usage_in_parent_managers(original_prediction, cost)


    @property
    def cost_per_token_completion(self) -> float:
        return self.caculate_cost_from_tokens(prompt_tkns=0, completion_tkns=1)


    @property
    def cost_per_token_prompt(self) -> float:
        return self.caculate_cost_from_tokens(prompt_tkns=1, completion_tkns=0)


from sl_ai_models.utils.openai_utils import OpenAiUtils
from tests.ai_mock_manager import AiModelMockManager

################################## Message Creation Tests ##################################
def test_user_message_creator_has_only_one_message():
    messages = OpenAiUtils.put_single_user_message_in_list_using_prompt("Hello")
    length_of_messages = len(messages)
    assert length_of_messages == 1, "Length of user message from propmt is not 1"


def test_system_and_user_message_creator_has_two_messages():
    messages = OpenAiUtils.create_sytem_and_user_message_from_prompt("Hello", "Hi")
    length_of_messages = len(messages)
    assert length_of_messages == 2, "Length of system and user message from propmt is not 2"


def test_vision_message_creator_has_one_message():
    vision_data = AiModelMockManager.CHEAP_IMAGE_MESSAGE_DATA
    messages = OpenAiUtils.put_single_image_message_in_list_using_gpt_vision_input(vision_data)
    length_of_messages = len(messages)
    assert length_of_messages == 1, "Length of vision message from propmt is not 1"


def test_system_and_vision_message_creator_has_two_messages():
    vision_data = AiModelMockManager.CHEAP_IMAGE_MESSAGE_DATA
    messages = OpenAiUtils.create_system_and_image_message_from_prompt(vision_data, "Hi")
    length_of_messages = len(messages)
    assert length_of_messages == 2, "Length of system and vision message from propmt is not 2"


from __future__ import annotations
import logging
from logging.handlers import RotatingFileHandler
from logging import StreamHandler
from logging import FileHandler
import os
from sl_ai_models.utils import file_manipulation
import sys
import datetime


class CustomLogger:
    _initialized = False
    DEFAULT_MESSAGE_FORMAT = '%(threadName)s - %(asctime)s - %(levelname)s - %(name)s - %(funcName)s  - %(message)s'
    LATEST_DEBUG_LOG_FILE_PATH = file_manipulation.get_absolute_path("logs/latest_debug.log")
    ERROR_LOG_FILE_PATH = file_manipulation.get_absolute_path("logs/warnings/warnings.log")
    DEBUG_LOG_FILE_PATH = file_manipulation.get_absolute_path("logs/debug/debug.log")
    INFO_LOG_FILE_PATH = file_manipulation.get_absolute_path("logs/info/info.log")
    LATEST_INFO_LOG_FILE_PATH = file_manipulation.get_absolute_path("logs/latest_info.log")
    __message_to_append_to_file = "Message to be set..."


    @classmethod
    def set_up_logging(cls) -> None:
        if cls._initialized:
            return

        root_logger = logging.getLogger()
        root_logger.setLevel(logging.DEBUG)
        handlers = []

        file_writing_is_allowed = (os.environ.get("FILE_WRITING_ALLOWED", "TRUE").upper() == "TRUE")
        if file_writing_is_allowed:
            cls.__message_to_append_to_file = f"Root Logger initialized at {datetime.datetime.now()}\n"
            handler_1 = cls.create_persistent_log_file_handler(logging.WARNING, cls.ERROR_LOG_FILE_PATH)
            handler_2 = cls.create_persistent_log_file_handler(logging.INFO, cls.INFO_LOG_FILE_PATH)
            handler_3 = cls.create_persistent_log_file_handler(logging.DEBUG, cls.DEBUG_LOG_FILE_PATH)
            handler_4 = cls.create_latest_log_file_handler(logging.DEBUG, cls.LATEST_DEBUG_LOG_FILE_PATH)
            handler_5 = cls.create_latest_log_file_handler(logging.INFO, cls.LATEST_INFO_LOG_FILE_PATH)
            handlers.extend([handler_1, handler_2, handler_3, handler_4, handler_5])

        handler_6 = cls.create_stream_handler(logging.DEBUG)
        handlers.append(handler_6)

        for handler in handlers:
            root_logger.addHandler(handler)

        cls._clear_latest_log_files()
        root_logger.info(f"Logger initialized with {len(handlers)} handlers at {datetime.datetime.now()}")

        cls._initialized = True


    @classmethod
    def create_persistent_log_file_handler(cls, log_level: int, file_path: str) -> RotatingFileHandler:
        file_manipulation.create_or_append_to_file(file_path, cls.__message_to_append_to_file)
        formatter = logging.Formatter(cls.DEFAULT_MESSAGE_FORMAT)
        handler = RotatingFileHandler(file_path, maxBytes=1000000, backupCount=10)
        handler.setLevel(log_level)
        handler.setFormatter(formatter)
        return handler


    @classmethod
    def create_latest_log_file_handler(cls, log_level: int, file_path: str) -> FileHandler:
        file_manipulation.create_or_append_to_file(file_path, cls.__message_to_append_to_file)
        formatter = logging.Formatter(cls.DEFAULT_MESSAGE_FORMAT)
        handler = FileHandler(file_path)
        handler.setLevel(log_level)
        handler.setFormatter(formatter)
        return handler


    @classmethod
    def create_stream_handler(cls, log_level: int) -> StreamHandler:
        formatter = logging.Formatter(cls.DEFAULT_MESSAGE_FORMAT)
        handler = StreamHandler(sys.stdout)
        handler.setLevel(log_level)
        handler.setFormatter(formatter)
        return handler

    @classmethod
    def _clear_latest_log_files(cls) -> None:
        file_manipulation.create_or_overwrite_file(cls.LATEST_DEBUG_LOG_FILE_PATH, "")
        file_manipulation.create_or_overwrite_file(cls.LATEST_INFO_LOG_FILE_PATH, "")





from datetime import datetime, timedelta
import logging
logger = logging.getLogger(__name__)
import asyncio
from typing import Final
from enum import Enum


class LimitReachedResponse(Enum):
    RAISE_EXCEPTION = 1
    WAIT = 2


class ResourceUnavailableError(RuntimeError):
    """Raised when resources are unavailable and cannot continue execution."""


class ResourceUseEntry:
    def __init__(self, resources_used: int, time: datetime) -> None:
        self.resources_used: int = resources_used
        self.time: datetime = time


class RefreshingBucketRateLimiter:
    """
    The refreshing bucket rate limiter is a way of limiting resource use over time.

    For example:
    - requests per minute
    - tokens per second
    - etc.

    It sets up a capacity and a refresh rate.
    The capacity acts as a burst limit. You can spend X resources in a short interval before being forced to slow down.
    If you use only X capacity, it will take the (refresh rate * X) to get back to full capacity.
    If you reach the bottom of the bucket, the bucket will fill all the way up before you can use resources again.
    This is to make sure something like a "requests per minute" limit is not exceeded even after a burst
    (since averaging out the burst over the full recharge period would successfully hold to the limit).
    """


    def __init__(self, capacity: float, refresh_rate: float, limit_reached_response: LimitReachedResponse = LimitReachedResponse.WAIT) -> None:
        if capacity <= 0:
            raise ValueError("capacity must be greater than 0")
        self.capacity: Final[float] = capacity

        if refresh_rate < 0:
            raise ValueError("refresh_rate must not be negative")
        elif refresh_rate == 0:
            logger.info("refresh_rate is 0, resources will not refresh")
        self.refresh_rate: Final[float] = refresh_rate

        self.__limit_reached_response: LimitReachedResponse = limit_reached_response
        self.__available_resources: float = capacity
        self.__resource_history: list[ResourceUseEntry] = []
        self.__last_replenish_time: datetime = datetime.now()
        self.__available_resource_lock = asyncio.Lock()
        self.__resource_history_lock = asyncio.Lock()
        self.__fill_the_bucket_mode = False


    def refresh_and_then_get_available_resources(self) -> float:
        asyncio.run(self._refresh_resource_count())
        return self._available_resources


    def zero_out_resources(self) -> None:
        self._available_resources = 0
        asyncio.run(self.__update_fill_the_bucket_mode(resources_ran_out=True))


    @property
    def _available_resources(self) -> float:
        return self.__available_resources


    @_available_resources.setter
    def _available_resources(self, value: float) -> None:
        if value > self.capacity:
            raise ValueError("value must be less than or equal to capacity")
        elif value < 0:
            raise ValueError("value must be greater than or equal to 0")
        self.__available_resources = value


    def caculate_resources_passed_into_acquire_in_time_range(self, start_time: datetime, end_time: datetime) -> int:
        resources_used = 0
        for entry in self.__resource_history:
            if start_time < entry.time < end_time:
                resources_used += entry.resources_used

        return resources_used


    async def wait_till_able_to_acquire_resources(self, resources_being_consumed: int) -> None:
        if resources_being_consumed > self.capacity:
            raise ValueError(f"resources_being_consumed must be less than or equal to capacity. Capacity: {self.capacity}, resources_being_consumed: {resources_being_consumed}")

        resources_are_available: bool = await self._determine_if_resources_available(resources_being_consumed)
        await self.__update_fill_the_bucket_mode(resources_ran_out=(not resources_are_available))

        if not resources_are_available and self.__limit_reached_response == LimitReachedResponse.RAISE_EXCEPTION:
            raise ResourceUnavailableError("Resources not available. Limit Reached Response is RAISE_EXCEPTION")

        if not resources_are_available and self.refresh_rate == 0:
            raise RuntimeError("Resources not available. Would have waited indefinitely. refresh_rate is 0")

        while not resources_are_available or self.__fill_the_bucket_mode:
            seconds_to_sleep = await self.__caculate_seconds_to_sleep(resources_being_consumed)
            await asyncio.sleep(seconds_to_sleep)
            resources_are_available = await self._determine_if_resources_available(resources_being_consumed)
            await self.__update_fill_the_bucket_mode(resources_ran_out=(not resources_are_available))

        async with self.__available_resource_lock:
            self._available_resources -= resources_being_consumed

        await self.__add_resource_use_entry(resources_being_consumed)




    async def _determine_if_resources_available(self, resources_being_consumed: int) -> bool:
        await self._refresh_resource_count()
        async with self.__available_resource_lock:
            resources_are_available = resources_being_consumed <= self._available_resources
            return resources_are_available


    async def __update_fill_the_bucket_mode(self, resources_ran_out: bool) -> None:
        await self._refresh_resource_count()
        if self._available_resources >= self.capacity:
            if resources_ran_out == True:
                raise ValueError("resources_ran_out should not be True when resources are available")
            self.__fill_the_bucket_mode = False
        if resources_ran_out:
            self.__fill_the_bucket_mode = True


    async def _refresh_resource_count(self) -> None:
        async with self.__available_resource_lock:
            time_since_last_replenish: timedelta = datetime.now() - self.__last_replenish_time
            seconds_since_last_replenish: float = time_since_last_replenish.total_seconds()
            replenish_amount = seconds_since_last_replenish * self.refresh_rate
            new_total = self._available_resources + replenish_amount
            self._available_resources = min(new_total, self.capacity)
            self.__last_replenish_time = datetime.now()


    async def __caculate_seconds_to_sleep(self, resources_being_consumed: int) -> float:
        if self.__fill_the_bucket_mode:
            seconds_till_bucket_is_full = (self.capacity - self._available_resources) / self.refresh_rate
            return seconds_till_bucket_is_full
        elif resources_being_consumed > self._available_resources:
            seconds_till_resources_available = (resources_being_consumed - self._available_resources) / self.refresh_rate
            return seconds_till_resources_available
        else:
            return 0


    async def __add_resource_use_entry(self, resource_amount: int) -> None:
        async with self.__resource_history_lock:
            new_entry = ResourceUseEntry(resource_amount, datetime.now())
            self.__resource_history.append(new_entry)





from __future__ import annotations
from abc import ABC
from typing import Any
from sl_ai_models.basic_model_interfaces.ai_model import AiModel
from typing import Callable, TypeVar, Coroutine
import logging
logger = logging.getLogger(__name__)
import functools
from tenacity import retry, wait_random_exponential, stop_after_attempt


T = TypeVar('T')



class RetryableModel(AiModel, ABC):
    _DEFAULT_ALLOWED_TRIES: int = 3

    def __init__(self, allowed_tries: int = _DEFAULT_ALLOWED_TRIES, **kwargs) -> None:
        super().__init__(**kwargs)
        if isinstance(allowed_tries, int) and allowed_tries > 0:
            self.allowed_tries = allowed_tries
        else:
            raise ValueError(f"allowed_tries must exist and be an integer greater than 0. allowed_tries was: {allowed_tries}")


    @property
    def allowed_tries(self) -> int:
        return self.__allowed_tries


    @allowed_tries.setter
    def allowed_tries(self, value: int) -> None:
        if not isinstance(value, int) or value < 1:
            raise ValueError("allowed_tries must be an integer greater than 0.")
        self.__allowed_tries = value


    @staticmethod
    def _retry_according_to_model_allowed_tries(func: Callable[..., Coroutine[Any, Any, T]]) -> Callable[..., Coroutine[Any, Any, T]]:
        @functools.wraps(func)
        async def wrapper_with_access_to_self_variable(self: RetryableModel, *args, **kwargs) -> T:
            @retry(
                wait=wait_random_exponential(exp_base=2, multiplier=10, min=5, max=60), # Waits random number between 0 and exp_base^current_attempt * multiplier (with min and max override as needed)
                reraise=True,
                stop=stop_after_attempt(self.allowed_tries)
            )
            async def wrapper_with_action(self: RetryableModel, *args, **kwargs) -> T:
                result = await func(self, *args, **kwargs)
                return result
            return await wrapper_with_action(self, *args, **kwargs)
        return wrapper_with_access_to_self_variable




from abc import ABC, abstractmethod
from typing import Any
import os
from contextvars import ContextVar


class AiModel(ABC):

    _num_calls_to_dependent_model: ContextVar[int] = ContextVar('_num_calls_to_dependent_model', default=0)

    @abstractmethod
    async def invoke(self, input: Any) -> Any:
        """
        This function is the main function that should be called to use the model.
        """
        pass


    @abstractmethod
    async def _mockable_direct_call_to_model(self, input: Any) -> Any:
        """
        This function calls the model directly.
        The function will probably be mocked in tests with a static return value.
        Do not put any logic that is tested into this function if the mocked return value would break the test.
        """



    @staticmethod
    @abstractmethod
    def _get_mock_return_for_direct_call_to_model_using_cheap_input() -> Any:
        """
        This is a value that tests can use to mock the
        return value of the function for the direct call to the model.
        """
        pass


    @staticmethod
    @abstractmethod
    def _get_cheap_input_for_invoke() -> Any:
        """
        This is an input to the invoke function that is cheap to run
        for the tests that run real requests to the model
        """
        pass


    def _everything_special_to_call_before_direct_call(self) -> Any:
        """
        This function is called before the direct call to the model.
        It is a place to put any logic that should be run before the direct call.
        """
        self._increment_calls_then_error_if_testing_call_limit_reached()
        self._deactivate_langchain_traces_if_in_testing_mode()


    def _increment_calls_then_error_if_testing_call_limit_reached(self, max_calls: int = 30) -> None:
        current_calls = self._num_calls_to_dependent_model.get()
        self._num_calls_to_dependent_model.set(current_calls + 1)
        if 'PYTEST_CURRENT_TEST' in os.environ and self._num_calls_to_dependent_model.get() > max_calls:
            raise RuntimeError(f"model called more than {max_calls} times during testing")


    def _deactivate_langchain_traces_if_in_testing_mode(self) -> None:
        if 'PYTEST_CURRENT_TEST' in os.environ:
            if 'LANGCHAIN_API_KEY' in os.environ:
                del os.environ['LANGCHAIN_API_KEY']
            if 'LANGCHAIN_TRACING_V2' in os.environ:
                del os.environ['LANGCHAIN_TRACING_V2']






from __future__ import annotations
from abc import ABC
from typing import Any


class NamedModel(ABC):
    MODEL_NAME: str


    def __init_subclass__(cls: type[NamedModel], **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        if ABC not in cls.__bases__ and cls.MODEL_NAME is NotImplemented:
            raise NotImplementedError('You forgot to define MODEL_NAME')
